// Copyright 2023 The Chromium Authors
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "extensions/browser/browser_context_data.h"

#include <memory>

#include "content/public/browser/render_frame_host.h"
#include "content/public/browser/web_contents.h"

namespace extensions {

std::unique_ptr<ContextData> BrowserContextData::Clone() const {
  return std::make_unique<BrowserContextData>(frame_);
}

std::unique_ptr<ContextData> BrowserContextData::GetLocalParentOrOpener()
    const {
  content::RenderFrameHost* parent_or_opener = frame_->GetParent();
  // Non primary pages(e.g. fenced frame, prerendered page, bfcache, and
  // portals) can't look at the opener, and WebContents::GetOpener returns the
  // opener on the primary frame tree. Thus, GetOpener should be called when
  // |frame_| is a primary main frame.
  if (!parent_or_opener && frame_->IsInPrimaryMainFrame()) {
    parent_or_opener =
        content::WebContents::FromRenderFrameHost(frame_)->GetOpener();
  }
  if (!parent_or_opener) {
    return nullptr;
  }

  // Renderer-side WebLocalFrameAdapter only considers local frames.
  // Comparing processes is robust way to replicate such renderer-side checks,
  // because out caller (DoesContentScriptMatch) accepts false positives.
  // This comparison might be less accurate (e.g. give more false positives)
  // than SiteInstance comparison, but comparing processes should be robust
  // and stable as SiteInstanceGroup refactoring proceeds.
  if (parent_or_opener->GetProcess() != frame_->GetProcess()) {
    return nullptr;
  }

  return std::make_unique<BrowserContextData>(parent_or_opener);
}

GURL BrowserContextData::GetUrl() const {
  if (frame_->GetLastCommittedURL().is_empty()) {
    // It's possible for URL to be empty when `frame_` is on the initial empty
    // document. TODO(https://crbug.com/1197308): Consider making  `frame_`'s
    // document's URL about:blank instead of empty in that case.
    return GURL(url::kAboutBlankURL);
  }
  return frame_->GetLastCommittedURL();
}

url::Origin BrowserContextData::GetOrigin() const {
  return frame_->GetLastCommittedOrigin();
}

// BrowserContextData::CanAccess is unable to replicate all of the
// WebSecurityOrigin::CanAccess checks, so these methods should not be called.
bool BrowserContextData::CanAccess(const url::Origin& target) const {
  NOTREACHED();
  return true;
}

bool BrowserContextData::CanAccess(const ContextData& target) const {
  NOTREACHED();
  return true;
}

uintptr_t BrowserContextData::GetId() const {
  return frame_->GetRoutingID();
}

}  // namespace extensions
